/*
    This file is part of magpar.

    Copyright (C) 2002-2008 Werner Scholz

    www:   http://magnet.atp.tuwien.ac.at/scholz/magpar/
    email: magpar(at)magnet.atp.tuwien.ac.at

    magpar is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    magpar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with magpar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

static char const Id[] = "$Id: area.c 1638 2008-08-01 13:05:39Z scholz $\n\n";
static char const Td[] = "$Today: " __FILE__ "  " __DATE__ "  "  __TIME__  " $\n\n";

#include "griddata.h"

int AreaCal(Mat AIpol, PetscReal *p_area)
{
  PetscReal area;

  int ngrows,ngcols;
  int nlrows,nlcols;

  Vec       test1,test2;
  PetscReal vmin,vmax;
  int       imin,imax;

  MagparFunctionLogBegin;

  ierr = MatGetSize(AIpol,&ngrows,&ngcols);CHKERRQ(ierr);
  ierr = MatGetLocalSize(AIpol,&nlrows,&nlcols);CHKERRQ(ierr);

  ierr = VecCreate(PETSC_COMM_WORLD,&test1);CHKERRQ(ierr);
  ierr = VecCreate(PETSC_COMM_WORLD,&test2);CHKERRQ(ierr);
  ierr = VecSetSizes(test1,nlcols,ngcols);CHKERRQ(ierr);
  ierr = VecSetSizes(test2,nlrows,ngrows);CHKERRQ(ierr);
  ierr = VecSetFromOptions(test1);CHKERRQ(ierr);
  ierr = VecSetFromOptions(test2);CHKERRQ(ierr);

#if PETSC_VERSION >= 230
  ierr = VecSet(test1,1.0);CHKERRQ(ierr);
#else
  const PetscReal c_one=1;
  ierr = VecSet(&c_one,test1);CHKERRQ(ierr);
#endif

  ierr = MatMult(AIpol,test1,test2);CHKERRQ(ierr);

  /* check result of interpolation */
  ierr = VecMin(test2,&imin,&vmin);CHKERRQ(ierr);
  ierr = VecMax(test2,&imax,&vmax);CHKERRQ(ierr);

  if (vmin<0.0 || vmax>1.0+D_EPS) {
    PetscPrintf(PETSC_COMM_WORLD,
      "need to rescale matrix: %i: %e  %i: %e\n",
      imin,vmin,imax,vmax
    );

    /* TODO: rescale AIpol (not sure why this is necessary)
       seems like some pixels belong to two elements (just on interface!?)
    */
    PetscReal *atest2;
    ierr = VecGetArray(test2,&atest2);CHKERRQ(ierr);
    for (int i=0;i<nlrows;i++) {
      if (atest2[i]>1.0) atest2[i]=1.0/atest2[i];
    }
    ierr = VecRestoreArray(test2,&atest2);CHKERRQ(ierr);

    ierr = MatDiagonalScale(AIpol,test2,PETSC_NULL);CHKERRQ(ierr);

    ierr = MatMult(AIpol,test1,test2);CHKERRQ(ierr);
  }

  /* double check result of interpolation */
  ierr = VecMin(test2,&imin,&vmin);CHKERRQ(ierr);
  ierr = VecMax(test2,&imax,&vmax);CHKERRQ(ierr);

  PetscPrintf(PETSC_COMM_WORLD,
    "min/max: %i: %e  %i: %e\n",
    imin,vmin,imax,vmax
  );

  assert(vmin>=0.0);
  assert(vmax<=1.0+D_EPS);

  /* area calculation */
  ierr = VecSum(test2,&area);CHKERRQ(ierr);
  PetscPrintf(PETSC_COMM_WORLD,"area = %g/%i=%g%%\n",area,ngrows,area/ngrows*100);

  *p_area=area;

  ierr = VecDestroy(test1);
  ierr = VecDestroy(test2);

  MagparFunctionLogReturn(0);
}

