/*
    This file is part of magpar.

    Copyright (C) 2002-2008 Werner Scholz

    www:   http://magnet.atp.tuwien.ac.at/scholz/magpar/
    email: magpar(at)magnet.atp.tuwien.ac.at

    magpar is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    magpar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with magpar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

static char const Id[] = "$Id: distortvec.c 2313 2008-12-02 04:34:38Z scholz $\n\n";
static char const Td[] = "$Today: " __FILE__ "  " __DATE__ "  "  __TIME__  " $\n\n";

#include "griddata.h"

int DistortVec(PetscRandom rctx, Vec v, PetscReal distpar)
{
  MagparFunctionInfoBegin;

  Vec VecRnd;
  ierr = VecDuplicate(v,&VecRnd);CHKERRQ(ierr);

#if PETSC_VERSION >= 230
  ierr = VecSetRandom(VecRnd,rctx);CHKERRQ(ierr);

  ierr = VecScale(VecRnd,2.0);CHKERRQ(ierr);
  ierr = VecShift(VecRnd,-1.0);CHKERRQ(ierr);

  ierr = VecAXPY(v,distpar,VecRnd);CHKERRQ(ierr);
#else
  ierr = VecSetRandom(rctx,VecRnd);CHKERRQ(ierr);

  PetscReal mone=-1.0, two=2.0;
  ierr = VecScale(&two,VecRnd);CHKERRQ(ierr);
  ierr = VecShift(&mone,VecRnd);CHKERRQ(ierr);

  ierr = VecAXPY(&distpar,VecRnd,v);CHKERRQ(ierr);
#endif

  ierr = VecDestroy(VecRnd);CHKERRQ(ierr);

  MagparFunctionInfoReturn(0);
}

